/*
 * Copyright (C) 2010-2018 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include <WebCore/DOMCacheEngine.h>
#include <wtf/Forward.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace PAL {
class SessionID;
}

namespace WebCore {
class ResourceRequest;
struct ClientOrigin;
struct CacheQueryOptions;
}

namespace Messages {
namespace CacheStorageEngineConnection {

static inline IPC::StringReference messageReceiverName()
{
    return IPC::StringReference("CacheStorageEngineConnection");
}

class Reference {
public:
    typedef std::tuple<const PAL::SessionID&, uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("Reference"); }
    static const bool isSync = false;

    Reference(const PAL::SessionID& sessionID, uint64_t cacheIdentifier)
        : m_arguments(sessionID, cacheIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Dereference {
public:
    typedef std::tuple<const PAL::SessionID&, uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("Dereference"); }
    static const bool isSync = false;

    Dereference(const PAL::SessionID& sessionID, uint64_t cacheIdentifier)
        : m_arguments(sessionID, cacheIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Open {
public:
    typedef std::tuple<const PAL::SessionID&, uint64_t, const WebCore::ClientOrigin&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("Open"); }
    static const bool isSync = false;

    Open(const PAL::SessionID& sessionID, uint64_t requestIdentifier, const WebCore::ClientOrigin& origin, const String& cacheName)
        : m_arguments(sessionID, requestIdentifier, origin, cacheName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Remove {
public:
    typedef std::tuple<const PAL::SessionID&, uint64_t, uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("Remove"); }
    static const bool isSync = false;

    Remove(const PAL::SessionID& sessionID, uint64_t requestIdentifier, uint64_t cacheIdentifier)
        : m_arguments(sessionID, requestIdentifier, cacheIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Caches {
public:
    typedef std::tuple<const PAL::SessionID&, uint64_t, const WebCore::ClientOrigin&, uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("Caches"); }
    static const bool isSync = false;

    Caches(const PAL::SessionID& sessionID, uint64_t requestIdentifier, const WebCore::ClientOrigin& origin, uint64_t updateCounter)
        : m_arguments(sessionID, requestIdentifier, origin, updateCounter)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ClearMemoryRepresentation {
public:
    typedef std::tuple<const PAL::SessionID&, uint64_t, const WebCore::ClientOrigin&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ClearMemoryRepresentation"); }
    static const bool isSync = false;

    ClearMemoryRepresentation(const PAL::SessionID& sessionID, uint64_t requestIdentifier, const WebCore::ClientOrigin& origin)
        : m_arguments(sessionID, requestIdentifier, origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class EngineRepresentation {
public:
    typedef std::tuple<const PAL::SessionID&, uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("EngineRepresentation"); }
    static const bool isSync = false;

    EngineRepresentation(const PAL::SessionID& sessionID, uint64_t requestIdentifier)
        : m_arguments(sessionID, requestIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RetrieveRecords {
public:
    typedef std::tuple<const PAL::SessionID&, uint64_t, uint64_t, const URL&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RetrieveRecords"); }
    static const bool isSync = false;

    RetrieveRecords(const PAL::SessionID& sessionID, uint64_t requestIdentifier, uint64_t cacheIdentifier, const URL& url)
        : m_arguments(sessionID, requestIdentifier, cacheIdentifier, url)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DeleteMatchingRecords {
public:
    typedef std::tuple<const PAL::SessionID&, uint64_t, uint64_t, const WebCore::ResourceRequest&, const WebCore::CacheQueryOptions&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DeleteMatchingRecords"); }
    static const bool isSync = false;

    DeleteMatchingRecords(const PAL::SessionID& sessionID, uint64_t requestIdentifier, uint64_t cacheIdentifier, const WebCore::ResourceRequest& request, const WebCore::CacheQueryOptions& options)
        : m_arguments(sessionID, requestIdentifier, cacheIdentifier, request, options)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PutRecords {
public:
    typedef std::tuple<const PAL::SessionID&, uint64_t, uint64_t, const Vector<WebCore::DOMCacheEngine::Record>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("PutRecords"); }
    static const bool isSync = false;

    PutRecords(const PAL::SessionID& sessionID, uint64_t requestIdentifier, uint64_t cacheIdentifier, const Vector<WebCore::DOMCacheEngine::Record>& record)
        : m_arguments(sessionID, requestIdentifier, cacheIdentifier, record)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

} // namespace CacheStorageEngineConnection
} // namespace Messages
