/*
 * Copyright (C) 2010-2018 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#if ENABLE(NETSCAPE_PLUGIN_API)

#include "NPObjectMessageReceiver.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "NPIdentifierData.h"
#include "NPObjectMessageReceiverMessages.h"
#include "NPVariantData.h"
#include <wtf/Vector.h>

namespace Messages {

namespace NPObjectMessageReceiver {

void Deallocate::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

void HasMethod::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool returnValue)
{
    *encoder << returnValue;
    connection.sendSyncReply(WTFMove(encoder));
}

void Invoke::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool returnValue, const WebKit::NPVariantData& resultData)
{
    *encoder << returnValue;
    *encoder << resultData;
    connection.sendSyncReply(WTFMove(encoder));
}

void InvokeDefault::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool returnValue, const WebKit::NPVariantData& resultData)
{
    *encoder << returnValue;
    *encoder << resultData;
    connection.sendSyncReply(WTFMove(encoder));
}

void HasProperty::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool returnValue)
{
    *encoder << returnValue;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetProperty::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool returnValue, const WebKit::NPVariantData& resultData)
{
    *encoder << returnValue;
    *encoder << resultData;
    connection.sendSyncReply(WTFMove(encoder));
}

void SetProperty::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool returnValue)
{
    *encoder << returnValue;
    connection.sendSyncReply(WTFMove(encoder));
}

void RemoveProperty::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool returnValue)
{
    *encoder << returnValue;
    connection.sendSyncReply(WTFMove(encoder));
}

void Enumerate::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool returnValue, const Vector<WebKit::NPIdentifierData>& identifiersData)
{
    *encoder << returnValue;
    *encoder << identifiersData;
    connection.sendSyncReply(WTFMove(encoder));
}

void Construct::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool returnValue, const WebKit::NPVariantData& resultData)
{
    *encoder << returnValue;
    *encoder << resultData;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace NPObjectMessageReceiver

} // namespace Messages

namespace WebKit {


void NPObjectMessageReceiver::didReceiveSyncNPObjectMessageReceiverMessage(IPC::Connection& connection, IPC::Decoder& decoder, std::unique_ptr<IPC::Encoder>& replyEncoder)
{
    if (decoder.messageName() == Messages::NPObjectMessageReceiver::Deallocate::name()) {
        IPC::handleMessageDelayed<Messages::NPObjectMessageReceiver::Deallocate>(connection, decoder, replyEncoder, this, &NPObjectMessageReceiver::deallocate);
        return;
    }
    if (decoder.messageName() == Messages::NPObjectMessageReceiver::HasMethod::name()) {
        IPC::handleMessageDelayed<Messages::NPObjectMessageReceiver::HasMethod>(connection, decoder, replyEncoder, this, &NPObjectMessageReceiver::hasMethod);
        return;
    }
    if (decoder.messageName() == Messages::NPObjectMessageReceiver::Invoke::name()) {
        IPC::handleMessageDelayed<Messages::NPObjectMessageReceiver::Invoke>(connection, decoder, replyEncoder, this, &NPObjectMessageReceiver::invoke);
        return;
    }
    if (decoder.messageName() == Messages::NPObjectMessageReceiver::InvokeDefault::name()) {
        IPC::handleMessageDelayed<Messages::NPObjectMessageReceiver::InvokeDefault>(connection, decoder, replyEncoder, this, &NPObjectMessageReceiver::invokeDefault);
        return;
    }
    if (decoder.messageName() == Messages::NPObjectMessageReceiver::HasProperty::name()) {
        IPC::handleMessageDelayed<Messages::NPObjectMessageReceiver::HasProperty>(connection, decoder, replyEncoder, this, &NPObjectMessageReceiver::hasProperty);
        return;
    }
    if (decoder.messageName() == Messages::NPObjectMessageReceiver::GetProperty::name()) {
        IPC::handleMessageDelayed<Messages::NPObjectMessageReceiver::GetProperty>(connection, decoder, replyEncoder, this, &NPObjectMessageReceiver::getProperty);
        return;
    }
    if (decoder.messageName() == Messages::NPObjectMessageReceiver::SetProperty::name()) {
        IPC::handleMessageDelayed<Messages::NPObjectMessageReceiver::SetProperty>(connection, decoder, replyEncoder, this, &NPObjectMessageReceiver::setProperty);
        return;
    }
    if (decoder.messageName() == Messages::NPObjectMessageReceiver::RemoveProperty::name()) {
        IPC::handleMessageDelayed<Messages::NPObjectMessageReceiver::RemoveProperty>(connection, decoder, replyEncoder, this, &NPObjectMessageReceiver::removeProperty);
        return;
    }
    if (decoder.messageName() == Messages::NPObjectMessageReceiver::Enumerate::name()) {
        IPC::handleMessageDelayed<Messages::NPObjectMessageReceiver::Enumerate>(connection, decoder, replyEncoder, this, &NPObjectMessageReceiver::enumerate);
        return;
    }
    if (decoder.messageName() == Messages::NPObjectMessageReceiver::Construct::name()) {
        IPC::handleMessageDelayed<Messages::NPObjectMessageReceiver::Construct>(connection, decoder, replyEncoder, this, &NPObjectMessageReceiver::construct);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit


#endif // ENABLE(NETSCAPE_PLUGIN_API)
