/*
 * Copyright (C) 2010-2018 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include "NetworkProcess.h"

#include "ArgumentCoders.h"
#include "CacheModel.h"
#include "DataReference.h"
#include "Decoder.h"
#include "DownloadID.h"
#if ENABLE(SERVICE_WORKER)
#include "FormDataReference.h"
#endif
#include "HandleMessage.h"
#include "NetworkProcessCreationParameters.h"
#include "NetworkProcessMessages.h"
#include "SandboxExtension.h"
#if ENABLE(SERVICE_WORKER)
#include "SharedBufferDataReference.h"
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
#include "ShouldGrandfatherStatistics.h"
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
#include "StorageAccessStatus.h"
#endif
#include "WebCoreArgumentCoders.h"
#include "WebsiteDataFetchOption.h"
#include "WebsiteDataStoreParameters.h"
#include "WebsiteDataType.h"
#include <WebCore/CertificateInfo.h>
#if USE(CURL)
#include <WebCore/CurlProxySettings.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/FetchIdentifier.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/MessageWithMessagePorts.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ResourceError.h>
#endif
#include <WebCore/ResourceRequest.h>
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ResourceResponse.h>
#endif
#include <WebCore/SecurityOriginData.h>
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ServiceWorkerClientIdentifier.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ServiceWorkerIdentifier.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ServiceWorkerTypes.h>
#endif
#if USE(SOUP)
#include <WebCore/SoupNetworkProxySettings.h>
#endif
#include <WebCore/StoredCredentialsPolicy.h>
#include <pal/SessionID.h>
#include <wtf/OptionSet.h>
#if ENABLE(RESOURCE_LOAD_STATISTICS)
#include <wtf/Optional.h>
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
#include <wtf/Seconds.h>
#endif
#include <wtf/Vector.h>
#include <wtf/WallTime.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace NetworkProcess {

void SetAllowsAnySSLCertificateForWebSocket::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

void ProcessWillSuspendImminently::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool handled)
{
    *encoder << handled;
    connection.sendSyncReply(WTFMove(encoder));
}

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void ClearPrevalentResource::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ClearPrevalentResource::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ClearPrevalentResource::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void ClearUserInteraction::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ClearUserInteraction::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ClearUserInteraction::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void DumpResourceLoadStatistics::callReply(IPC::Decoder& decoder, CompletionHandler<void(String&&)>&& completionHandler)
{
    Optional<String> dumpedStatistics;
    decoder >> dumpedStatistics;
    if (!dumpedStatistics) {
        ASSERT_NOT_REACHED();
        return;
    }
    completionHandler(WTFMove(*dumpedStatistics));
}

void DumpResourceLoadStatistics::cancelReply(CompletionHandler<void(String&&)>&& completionHandler)
{
    completionHandler({ });
}

void DumpResourceLoadStatistics::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const String& dumpedStatistics)
{
    *encoder << dumpedStatistics;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void UpdatePrevalentDomainsToBlockCookiesFor::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void UpdatePrevalentDomainsToBlockCookiesFor::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void UpdatePrevalentDomainsToBlockCookiesFor::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void IsGrandfathered::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> isGrandfathered;
    decoder >> isGrandfathered;
    if (!isGrandfathered) {
        ASSERT_NOT_REACHED();
        return;
    }
    completionHandler(WTFMove(*isGrandfathered));
}

void IsGrandfathered::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler({ });
}

void IsGrandfathered::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool isGrandfathered)
{
    *encoder << isGrandfathered;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void IsPrevalentResource::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> isPrevalent;
    decoder >> isPrevalent;
    if (!isPrevalent) {
        ASSERT_NOT_REACHED();
        return;
    }
    completionHandler(WTFMove(*isPrevalent));
}

void IsPrevalentResource::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler({ });
}

void IsPrevalentResource::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool isPrevalent)
{
    *encoder << isPrevalent;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void IsVeryPrevalentResource::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> isVeryPrevalent;
    decoder >> isVeryPrevalent;
    if (!isVeryPrevalent) {
        ASSERT_NOT_REACHED();
        return;
    }
    completionHandler(WTFMove(*isVeryPrevalent));
}

void IsVeryPrevalentResource::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler({ });
}

void IsVeryPrevalentResource::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool isVeryPrevalent)
{
    *encoder << isVeryPrevalent;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetAgeCapForClientSideCookies::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetAgeCapForClientSideCookies::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetAgeCapForClientSideCookies::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetLastSeen::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetLastSeen::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetLastSeen::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetPrevalentResource::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetPrevalentResource::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetPrevalentResource::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetPrevalentResourceForDebugMode::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetPrevalentResourceForDebugMode::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetPrevalentResourceForDebugMode::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void HadUserInteraction::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> hadUserInteraction;
    decoder >> hadUserInteraction;
    if (!hadUserInteraction) {
        ASSERT_NOT_REACHED();
        return;
    }
    completionHandler(WTFMove(*hadUserInteraction));
}

void HadUserInteraction::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler({ });
}

void HadUserInteraction::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool hadUserInteraction)
{
    *encoder << hadUserInteraction;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void HasStorageAccessForFrame::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> hasStorageAccess;
    decoder >> hasStorageAccess;
    if (!hasStorageAccess) {
        ASSERT_NOT_REACHED();
        return;
    }
    completionHandler(WTFMove(*hasStorageAccess));
}

void HasStorageAccessForFrame::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler({ });
}

void HasStorageAccessForFrame::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool hasStorageAccess)
{
    *encoder << hasStorageAccess;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void HasStorageAccess::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> hasStorageAccess;
    decoder >> hasStorageAccess;
    if (!hasStorageAccess) {
        ASSERT_NOT_REACHED();
        return;
    }
    completionHandler(WTFMove(*hasStorageAccess));
}

void HasStorageAccess::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler({ });
}

void HasStorageAccess::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool hasStorageAccess)
{
    *encoder << hasStorageAccess;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void GetAllStorageAccessEntries::callReply(IPC::Decoder& decoder, CompletionHandler<void(Vector<String>&&)>&& completionHandler)
{
    Optional<Vector<String>> domains;
    decoder >> domains;
    if (!domains) {
        ASSERT_NOT_REACHED();
        return;
    }
    completionHandler(WTFMove(*domains));
}

void GetAllStorageAccessEntries::cancelReply(CompletionHandler<void(Vector<String>&&)>&& completionHandler)
{
    completionHandler({ });
}

void GetAllStorageAccessEntries::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<String>& domains)
{
    *encoder << domains;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void GrantStorageAccess::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> accessGranted;
    decoder >> accessGranted;
    if (!accessGranted) {
        ASSERT_NOT_REACHED();
        return;
    }
    completionHandler(WTFMove(*accessGranted));
}

void GrantStorageAccess::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler({ });
}

void GrantStorageAccess::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool accessGranted)
{
    *encoder << accessGranted;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void IsRegisteredAsRedirectingTo::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> isRedirectingTo;
    decoder >> isRedirectingTo;
    if (!isRedirectingTo) {
        ASSERT_NOT_REACHED();
        return;
    }
    completionHandler(WTFMove(*isRedirectingTo));
}

void IsRegisteredAsRedirectingTo::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler({ });
}

void IsRegisteredAsRedirectingTo::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool isRedirectingTo)
{
    *encoder << isRedirectingTo;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void IsRegisteredAsSubFrameUnder::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> isSubframeUnder;
    decoder >> isSubframeUnder;
    if (!isSubframeUnder) {
        ASSERT_NOT_REACHED();
        return;
    }
    completionHandler(WTFMove(*isSubframeUnder));
}

void IsRegisteredAsSubFrameUnder::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler({ });
}

void IsRegisteredAsSubFrameUnder::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool isSubframeUnder)
{
    *encoder << isSubframeUnder;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void IsRegisteredAsSubresourceUnder::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> isSubresourceUnder;
    decoder >> isSubresourceUnder;
    if (!isSubresourceUnder) {
        ASSERT_NOT_REACHED();
        return;
    }
    completionHandler(WTFMove(*isSubresourceUnder));
}

void IsRegisteredAsSubresourceUnder::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler({ });
}

void IsRegisteredAsSubresourceUnder::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool isSubresourceUnder)
{
    *encoder << isSubresourceUnder;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void LogUserInteraction::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void LogUserInteraction::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void LogUserInteraction::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void RemoveAllStorageAccess::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void RemoveAllStorageAccess::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void RemoveAllStorageAccess::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void RequestStorageAccess::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebKit::StorageAccessStatus&&)>&& completionHandler)
{
    Optional<WebKit::StorageAccessStatus> storageAccessStatus;
    decoder >> storageAccessStatus;
    if (!storageAccessStatus) {
        ASSERT_NOT_REACHED();
        return;
    }
    completionHandler(WTFMove(*storageAccessStatus));
}

void RequestStorageAccess::cancelReply(CompletionHandler<void(WebKit::StorageAccessStatus&&)>&& completionHandler)
{
    completionHandler({ });
}

void RequestStorageAccess::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, WebKit::StorageAccessStatus storageAccessStatus)
{
    *encoder << storageAccessStatus;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void ResetParametersToDefaultValues::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ResetParametersToDefaultValues::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ResetParametersToDefaultValues::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void ScheduleClearInMemoryAndPersistent::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ScheduleClearInMemoryAndPersistent::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ScheduleClearInMemoryAndPersistent::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void ScheduleCookieBlockingUpdate::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ScheduleCookieBlockingUpdate::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ScheduleCookieBlockingUpdate::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void ScheduleStatisticsAndDataRecordsProcessing::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ScheduleStatisticsAndDataRecordsProcessing::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ScheduleStatisticsAndDataRecordsProcessing::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SubmitTelemetry::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SubmitTelemetry::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SubmitTelemetry::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetCacheMaxAgeCapForPrevalentResources::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetCacheMaxAgeCapForPrevalentResources::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetCacheMaxAgeCapForPrevalentResources::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetGrandfathered::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetGrandfathered::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetGrandfathered::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetGrandfatheringTime::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetGrandfatheringTime::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetGrandfatheringTime::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetMaxStatisticsEntries::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetMaxStatisticsEntries::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetMaxStatisticsEntries::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetMinimumTimeBetweenDataRecordsRemoval::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetMinimumTimeBetweenDataRecordsRemoval::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetMinimumTimeBetweenDataRecordsRemoval::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetPruneEntriesDownTo::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetPruneEntriesDownTo::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetPruneEntriesDownTo::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetShouldClassifyResourcesBeforeDataRecordsRemoval::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetShouldClassifyResourcesBeforeDataRecordsRemoval::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetShouldClassifyResourcesBeforeDataRecordsRemoval::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetNotifyPagesWhenDataRecordsWereScanned::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetNotifyPagesWhenDataRecordsWereScanned::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetNotifyPagesWhenDataRecordsWereScanned::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetNotifyPagesWhenTelemetryWasCaptured::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetNotifyPagesWhenTelemetryWasCaptured::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetNotifyPagesWhenTelemetryWasCaptured::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetResourceLoadStatisticsDebugMode::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetResourceLoadStatisticsDebugMode::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetResourceLoadStatisticsDebugMode::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetVeryPrevalentResource::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetVeryPrevalentResource::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetVeryPrevalentResource::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetSubframeUnderTopFrameOrigin::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetSubframeUnderTopFrameOrigin::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetSubframeUnderTopFrameOrigin::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetSubresourceUnderTopFrameOrigin::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetSubresourceUnderTopFrameOrigin::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetSubresourceUnderTopFrameOrigin::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetSubresourceUniqueRedirectTo::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetSubresourceUniqueRedirectTo::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetSubresourceUniqueRedirectTo::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetSubresourceUniqueRedirectFrom::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetSubresourceUniqueRedirectFrom::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetSubresourceUniqueRedirectFrom::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetTimeToLiveUserInteraction::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetTimeToLiveUserInteraction::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetTimeToLiveUserInteraction::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetTopFrameUniqueRedirectTo::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetTopFrameUniqueRedirectTo::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetTopFrameUniqueRedirectTo::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetTopFrameUniqueRedirectFrom::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetTopFrameUniqueRedirectFrom::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetTopFrameUniqueRedirectFrom::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void ResetCacheMaxAgeCapForPrevalentResources::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ResetCacheMaxAgeCapForPrevalentResources::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ResetCacheMaxAgeCapForPrevalentResources::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

} // namespace NetworkProcess

} // namespace Messages

namespace WebKit {

void NetworkProcess::didReceiveNetworkProcessMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    if (decoder.messageName() == Messages::NetworkProcess::InitializeNetworkProcess::name()) {
        IPC::handleMessage<Messages::NetworkProcess::InitializeNetworkProcess>(decoder, this, &NetworkProcess::initializeNetworkProcess);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::CreateNetworkConnectionToWebProcess::name()) {
        IPC::handleMessage<Messages::NetworkProcess::CreateNetworkConnectionToWebProcess>(decoder, this, &NetworkProcess::createNetworkConnectionToWebProcess);
        return;
    }
#if USE(SOUP)
    if (decoder.messageName() == Messages::NetworkProcess::SetIgnoreTLSErrors::name()) {
        IPC::handleMessage<Messages::NetworkProcess::SetIgnoreTLSErrors>(decoder, this, &NetworkProcess::setIgnoreTLSErrors);
        return;
    }
#endif
#if USE(SOUP)
    if (decoder.messageName() == Messages::NetworkProcess::UserPreferredLanguagesChanged::name()) {
        IPC::handleMessage<Messages::NetworkProcess::UserPreferredLanguagesChanged>(decoder, this, &NetworkProcess::userPreferredLanguagesChanged);
        return;
    }
#endif
#if USE(SOUP)
    if (decoder.messageName() == Messages::NetworkProcess::SetNetworkProxySettings::name()) {
        IPC::handleMessage<Messages::NetworkProcess::SetNetworkProxySettings>(decoder, this, &NetworkProcess::setNetworkProxySettings);
        return;
    }
#endif
#if USE(CURL)
    if (decoder.messageName() == Messages::NetworkProcess::SetNetworkProxySettings::name()) {
        IPC::handleMessage<Messages::NetworkProcess::SetNetworkProxySettings>(decoder, this, &NetworkProcess::setNetworkProxySettings);
        return;
    }
#endif
    if (decoder.messageName() == Messages::NetworkProcess::ClearCachedCredentials::name()) {
        IPC::handleMessage<Messages::NetworkProcess::ClearCachedCredentials>(decoder, this, &NetworkProcess::clearCachedCredentials);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::AddWebsiteDataStore::name()) {
        IPC::handleMessage<Messages::NetworkProcess::AddWebsiteDataStore>(decoder, this, &NetworkProcess::addWebsiteDataStore);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::DestroySession::name()) {
        IPC::handleMessage<Messages::NetworkProcess::DestroySession>(decoder, this, &NetworkProcess::destroySession);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::FetchWebsiteData::name()) {
        IPC::handleMessage<Messages::NetworkProcess::FetchWebsiteData>(decoder, this, &NetworkProcess::fetchWebsiteData);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::DeleteWebsiteData::name()) {
        IPC::handleMessage<Messages::NetworkProcess::DeleteWebsiteData>(decoder, this, &NetworkProcess::deleteWebsiteData);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::DeleteWebsiteDataForOrigins::name()) {
        IPC::handleMessage<Messages::NetworkProcess::DeleteWebsiteDataForOrigins>(decoder, this, &NetworkProcess::deleteWebsiteDataForOrigins);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::DownloadRequest::name()) {
        IPC::handleMessage<Messages::NetworkProcess::DownloadRequest>(decoder, this, &NetworkProcess::downloadRequest);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::ResumeDownload::name()) {
        IPC::handleMessage<Messages::NetworkProcess::ResumeDownload>(decoder, this, &NetworkProcess::resumeDownload);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::CancelDownload::name()) {
        IPC::handleMessage<Messages::NetworkProcess::CancelDownload>(decoder, this, &NetworkProcess::cancelDownload);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::NetworkProcess::PublishDownloadProgress::name()) {
        IPC::handleMessage<Messages::NetworkProcess::PublishDownloadProgress>(decoder, this, &NetworkProcess::publishDownloadProgress);
        return;
    }
#endif
    if (decoder.messageName() == Messages::NetworkProcess::ContinueWillSendRequest::name()) {
        IPC::handleMessage<Messages::NetworkProcess::ContinueWillSendRequest>(decoder, this, &NetworkProcess::continueWillSendRequest);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::ContinueDecidePendingDownloadDestination::name()) {
        IPC::handleMessage<Messages::NetworkProcess::ContinueDecidePendingDownloadDestination>(decoder, this, &NetworkProcess::continueDecidePendingDownloadDestination);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::NetworkProcess::SetQOS::name()) {
        IPC::handleMessage<Messages::NetworkProcess::SetQOS>(decoder, this, &NetworkProcess::setQOS);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::NetworkProcess::SetStorageAccessAPIEnabled::name()) {
        IPC::handleMessage<Messages::NetworkProcess::SetStorageAccessAPIEnabled>(decoder, this, &NetworkProcess::setStorageAccessAPIEnabled);
        return;
    }
#endif
    if (decoder.messageName() == Messages::NetworkProcess::SyncAllCookies::name()) {
        IPC::handleMessage<Messages::NetworkProcess::SyncAllCookies>(decoder, this, &NetworkProcess::syncAllCookies);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::AllowSpecificHTTPSCertificateForHost::name()) {
        IPC::handleMessage<Messages::NetworkProcess::AllowSpecificHTTPSCertificateForHost>(decoder, this, &NetworkProcess::allowSpecificHTTPSCertificateForHost);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::SetCanHandleHTTPSServerTrustEvaluation::name()) {
        IPC::handleMessage<Messages::NetworkProcess::SetCanHandleHTTPSServerTrustEvaluation>(decoder, this, &NetworkProcess::setCanHandleHTTPSServerTrustEvaluation);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::GetNetworkProcessStatistics::name()) {
        IPC::handleMessage<Messages::NetworkProcess::GetNetworkProcessStatistics>(decoder, this, &NetworkProcess::getNetworkProcessStatistics);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::ClearCacheForAllOrigins::name()) {
        IPC::handleMessage<Messages::NetworkProcess::ClearCacheForAllOrigins>(decoder, this, &NetworkProcess::clearCacheForAllOrigins);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::SetCacheModel::name()) {
        IPC::handleMessage<Messages::NetworkProcess::SetCacheModel>(decoder, this, &NetworkProcess::setCacheModel);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::ProcessDidTransitionToBackground::name()) {
        IPC::handleMessage<Messages::NetworkProcess::ProcessDidTransitionToBackground>(decoder, this, &NetworkProcess::processDidTransitionToBackground);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::ProcessDidTransitionToForeground::name()) {
        IPC::handleMessage<Messages::NetworkProcess::ProcessDidTransitionToForeground>(decoder, this, &NetworkProcess::processDidTransitionToForeground);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::PrepareToSuspend::name()) {
        IPC::handleMessage<Messages::NetworkProcess::PrepareToSuspend>(decoder, this, &NetworkProcess::prepareToSuspend);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::CancelPrepareToSuspend::name()) {
        IPC::handleMessage<Messages::NetworkProcess::CancelPrepareToSuspend>(decoder, this, &NetworkProcess::cancelPrepareToSuspend);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::ProcessDidResume::name()) {
        IPC::handleMessage<Messages::NetworkProcess::ProcessDidResume>(decoder, this, &NetworkProcess::processDidResume);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::PreconnectTo::name()) {
        IPC::handleMessage<Messages::NetworkProcess::PreconnectTo>(decoder, this, &NetworkProcess::preconnectTo);
        return;
    }
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::ClearPrevalentResource::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::ClearPrevalentResource>(connection, decoder, this, &NetworkProcess::clearPrevalentResource);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::ClearUserInteraction::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::ClearUserInteraction>(connection, decoder, this, &NetworkProcess::clearUserInteraction);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::DumpResourceLoadStatistics::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::DumpResourceLoadStatistics>(connection, decoder, this, &NetworkProcess::dumpResourceLoadStatistics);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetResourceLoadStatisticsEnabled::name()) {
        IPC::handleMessage<Messages::NetworkProcess::SetResourceLoadStatisticsEnabled>(decoder, this, &NetworkProcess::setResourceLoadStatisticsEnabled);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::UpdatePrevalentDomainsToBlockCookiesFor::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::UpdatePrevalentDomainsToBlockCookiesFor>(connection, decoder, this, &NetworkProcess::updatePrevalentDomainsToBlockCookiesFor);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::IsGrandfathered::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::IsGrandfathered>(connection, decoder, this, &NetworkProcess::isGrandfathered);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::IsPrevalentResource::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::IsPrevalentResource>(connection, decoder, this, &NetworkProcess::isPrevalentResource);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::IsVeryPrevalentResource::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::IsVeryPrevalentResource>(connection, decoder, this, &NetworkProcess::isVeryPrevalentResource);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetAgeCapForClientSideCookies::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetAgeCapForClientSideCookies>(connection, decoder, this, &NetworkProcess::setAgeCapForClientSideCookies);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetLastSeen::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetLastSeen>(connection, decoder, this, &NetworkProcess::setLastSeen);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetPrevalentResource::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetPrevalentResource>(connection, decoder, this, &NetworkProcess::setPrevalentResource);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetPrevalentResourceForDebugMode::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetPrevalentResourceForDebugMode>(connection, decoder, this, &NetworkProcess::setPrevalentResourceForDebugMode);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::HadUserInteraction::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::HadUserInteraction>(connection, decoder, this, &NetworkProcess::hadUserInteraction);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::HasStorageAccessForFrame::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::HasStorageAccessForFrame>(connection, decoder, this, &NetworkProcess::hasStorageAccessForFrame);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::HasStorageAccess::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::HasStorageAccess>(connection, decoder, this, &NetworkProcess::hasStorageAccess);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::GetAllStorageAccessEntries::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::GetAllStorageAccessEntries>(connection, decoder, this, &NetworkProcess::getAllStorageAccessEntries);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::GrantStorageAccess::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::GrantStorageAccess>(connection, decoder, this, &NetworkProcess::grantStorageAccess);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::IsRegisteredAsRedirectingTo::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::IsRegisteredAsRedirectingTo>(connection, decoder, this, &NetworkProcess::isRegisteredAsRedirectingTo);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::IsRegisteredAsSubFrameUnder::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::IsRegisteredAsSubFrameUnder>(connection, decoder, this, &NetworkProcess::isRegisteredAsSubFrameUnder);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::IsRegisteredAsSubresourceUnder::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::IsRegisteredAsSubresourceUnder>(connection, decoder, this, &NetworkProcess::isRegisteredAsSubresourceUnder);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::LogFrameNavigation::name()) {
        IPC::handleMessage<Messages::NetworkProcess::LogFrameNavigation>(decoder, this, &NetworkProcess::logFrameNavigation);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::LogUserInteraction::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::LogUserInteraction>(connection, decoder, this, &NetworkProcess::logUserInteraction);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::RemoveAllStorageAccess::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::RemoveAllStorageAccess>(connection, decoder, this, &NetworkProcess::removeAllStorageAccess);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::RemovePrevalentDomains::name()) {
        IPC::handleMessage<Messages::NetworkProcess::RemovePrevalentDomains>(decoder, this, &NetworkProcess::removePrevalentDomains);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::RequestStorageAccess::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::RequestStorageAccess>(connection, decoder, this, &NetworkProcess::requestStorageAccess);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::ResetParametersToDefaultValues::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::ResetParametersToDefaultValues>(connection, decoder, this, &NetworkProcess::resetParametersToDefaultValues);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::ScheduleClearInMemoryAndPersistent::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::ScheduleClearInMemoryAndPersistent>(connection, decoder, this, &NetworkProcess::scheduleClearInMemoryAndPersistent);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::ScheduleCookieBlockingUpdate::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::ScheduleCookieBlockingUpdate>(connection, decoder, this, &NetworkProcess::scheduleCookieBlockingUpdate);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::ScheduleStatisticsAndDataRecordsProcessing::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::ScheduleStatisticsAndDataRecordsProcessing>(connection, decoder, this, &NetworkProcess::scheduleStatisticsAndDataRecordsProcessing);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SubmitTelemetry::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SubmitTelemetry>(connection, decoder, this, &NetworkProcess::submitTelemetry);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetCacheMaxAgeCapForPrevalentResources::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetCacheMaxAgeCapForPrevalentResources>(connection, decoder, this, &NetworkProcess::setCacheMaxAgeCapForPrevalentResources);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetGrandfathered::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetGrandfathered>(connection, decoder, this, &NetworkProcess::setGrandfathered);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetGrandfatheringTime::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetGrandfatheringTime>(connection, decoder, this, &NetworkProcess::setGrandfatheringTime);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetMaxStatisticsEntries::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetMaxStatisticsEntries>(connection, decoder, this, &NetworkProcess::setMaxStatisticsEntries);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetMinimumTimeBetweenDataRecordsRemoval::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetMinimumTimeBetweenDataRecordsRemoval>(connection, decoder, this, &NetworkProcess::setMinimumTimeBetweenDataRecordsRemoval);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetPruneEntriesDownTo::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetPruneEntriesDownTo>(connection, decoder, this, &NetworkProcess::setPruneEntriesDownTo);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetShouldClassifyResourcesBeforeDataRecordsRemoval::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetShouldClassifyResourcesBeforeDataRecordsRemoval>(connection, decoder, this, &NetworkProcess::setShouldClassifyResourcesBeforeDataRecordsRemoval);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetNotifyPagesWhenDataRecordsWereScanned::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetNotifyPagesWhenDataRecordsWereScanned>(connection, decoder, this, &NetworkProcess::setNotifyPagesWhenDataRecordsWereScanned);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetNotifyPagesWhenTelemetryWasCaptured::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetNotifyPagesWhenTelemetryWasCaptured>(connection, decoder, this, &NetworkProcess::setNotifyPagesWhenTelemetryWasCaptured);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetResourceLoadStatisticsDebugMode::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetResourceLoadStatisticsDebugMode>(connection, decoder, this, &NetworkProcess::setResourceLoadStatisticsDebugMode);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetVeryPrevalentResource::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetVeryPrevalentResource>(connection, decoder, this, &NetworkProcess::setVeryPrevalentResource);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetSubframeUnderTopFrameOrigin::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetSubframeUnderTopFrameOrigin>(connection, decoder, this, &NetworkProcess::setSubframeUnderTopFrameOrigin);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetSubresourceUnderTopFrameOrigin::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetSubresourceUnderTopFrameOrigin>(connection, decoder, this, &NetworkProcess::setSubresourceUnderTopFrameOrigin);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetSubresourceUniqueRedirectTo::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetSubresourceUniqueRedirectTo>(connection, decoder, this, &NetworkProcess::setSubresourceUniqueRedirectTo);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetSubresourceUniqueRedirectFrom::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetSubresourceUniqueRedirectFrom>(connection, decoder, this, &NetworkProcess::setSubresourceUniqueRedirectFrom);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetTimeToLiveUserInteraction::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetTimeToLiveUserInteraction>(connection, decoder, this, &NetworkProcess::setTimeToLiveUserInteraction);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetTopFrameUniqueRedirectTo::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetTopFrameUniqueRedirectTo>(connection, decoder, this, &NetworkProcess::setTopFrameUniqueRedirectTo);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::SetTopFrameUniqueRedirectFrom::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::SetTopFrameUniqueRedirectFrom>(connection, decoder, this, &NetworkProcess::setTopFrameUniqueRedirectFrom);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcess::ResetCacheMaxAgeCapForPrevalentResources::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcess::ResetCacheMaxAgeCapForPrevalentResources>(connection, decoder, this, &NetworkProcess::resetCacheMaxAgeCapForPrevalentResources);
        return;
    }
#endif
    if (decoder.messageName() == Messages::NetworkProcess::SetSessionIsControlledByAutomation::name()) {
        IPC::handleMessage<Messages::NetworkProcess::SetSessionIsControlledByAutomation>(decoder, this, &NetworkProcess::setSessionIsControlledByAutomation);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::RegisterURLSchemeAsSecure::name()) {
        IPC::handleMessage<Messages::NetworkProcess::RegisterURLSchemeAsSecure>(decoder, this, &NetworkProcess::registerURLSchemeAsSecure);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::RegisterURLSchemeAsBypassingContentSecurityPolicy::name()) {
        IPC::handleMessage<Messages::NetworkProcess::RegisterURLSchemeAsBypassingContentSecurityPolicy>(decoder, this, &NetworkProcess::registerURLSchemeAsBypassingContentSecurityPolicy);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::RegisterURLSchemeAsLocal::name()) {
        IPC::handleMessage<Messages::NetworkProcess::RegisterURLSchemeAsLocal>(decoder, this, &NetworkProcess::registerURLSchemeAsLocal);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::RegisterURLSchemeAsNoAccess::name()) {
        IPC::handleMessage<Messages::NetworkProcess::RegisterURLSchemeAsNoAccess>(decoder, this, &NetworkProcess::registerURLSchemeAsNoAccess);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::RegisterURLSchemeAsDisplayIsolated::name()) {
        IPC::handleMessage<Messages::NetworkProcess::RegisterURLSchemeAsDisplayIsolated>(decoder, this, &NetworkProcess::registerURLSchemeAsDisplayIsolated);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::RegisterURLSchemeAsCORSEnabled::name()) {
        IPC::handleMessage<Messages::NetworkProcess::RegisterURLSchemeAsCORSEnabled>(decoder, this, &NetworkProcess::registerURLSchemeAsCORSEnabled);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::RegisterURLSchemeAsCanDisplayOnlyIfCanRequest::name()) {
        IPC::handleMessage<Messages::NetworkProcess::RegisterURLSchemeAsCanDisplayOnlyIfCanRequest>(decoder, this, &NetworkProcess::registerURLSchemeAsCanDisplayOnlyIfCanRequest);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::SetCacheStorageParameters::name()) {
        IPC::handleMessage<Messages::NetworkProcess::SetCacheStorageParameters>(decoder, this, &NetworkProcess::setCacheStorageParameters);
        return;
    }
#if ENABLE(SERVICE_WORKER)
    if (decoder.messageName() == Messages::NetworkProcess::DidNotHandleFetch::name()) {
        IPC::handleMessage<Messages::NetworkProcess::DidNotHandleFetch>(decoder, this, &NetworkProcess::didNotHandleFetch);
        return;
    }
#endif
#if ENABLE(SERVICE_WORKER)
    if (decoder.messageName() == Messages::NetworkProcess::DidFailFetch::name()) {
        IPC::handleMessage<Messages::NetworkProcess::DidFailFetch>(decoder, this, &NetworkProcess::didFailFetch);
        return;
    }
#endif
#if ENABLE(SERVICE_WORKER)
    if (decoder.messageName() == Messages::NetworkProcess::DidReceiveFetchRedirectResponse::name()) {
        IPC::handleMessage<Messages::NetworkProcess::DidReceiveFetchRedirectResponse>(decoder, this, &NetworkProcess::didReceiveFetchRedirectResponse);
        return;
    }
#endif
#if ENABLE(SERVICE_WORKER)
    if (decoder.messageName() == Messages::NetworkProcess::DidReceiveFetchResponse::name()) {
        IPC::handleMessage<Messages::NetworkProcess::DidReceiveFetchResponse>(decoder, this, &NetworkProcess::didReceiveFetchResponse);
        return;
    }
#endif
#if ENABLE(SERVICE_WORKER)
    if (decoder.messageName() == Messages::NetworkProcess::DidReceiveFetchData::name()) {
        IPC::handleMessage<Messages::NetworkProcess::DidReceiveFetchData>(decoder, this, &NetworkProcess::didReceiveFetchData);
        return;
    }
#endif
#if ENABLE(SERVICE_WORKER)
    if (decoder.messageName() == Messages::NetworkProcess::DidReceiveFetchFormData::name()) {
        IPC::handleMessage<Messages::NetworkProcess::DidReceiveFetchFormData>(decoder, this, &NetworkProcess::didReceiveFetchFormData);
        return;
    }
#endif
#if ENABLE(SERVICE_WORKER)
    if (decoder.messageName() == Messages::NetworkProcess::DidFinishFetch::name()) {
        IPC::handleMessage<Messages::NetworkProcess::DidFinishFetch>(decoder, this, &NetworkProcess::didFinishFetch);
        return;
    }
#endif
#if ENABLE(SERVICE_WORKER)
    if (decoder.messageName() == Messages::NetworkProcess::PostMessageToServiceWorkerClient::name()) {
        IPC::handleMessage<Messages::NetworkProcess::PostMessageToServiceWorkerClient>(decoder, this, &NetworkProcess::postMessageToServiceWorkerClient);
        return;
    }
#endif
#if ENABLE(SERVICE_WORKER)
    if (decoder.messageName() == Messages::NetworkProcess::PostMessageToServiceWorker::name()) {
        IPC::handleMessage<Messages::NetworkProcess::PostMessageToServiceWorker>(decoder, this, &NetworkProcess::postMessageToServiceWorker);
        return;
    }
#endif
#if ENABLE(SERVICE_WORKER)
    if (decoder.messageName() == Messages::NetworkProcess::DisableServiceWorkerProcessTerminationDelay::name()) {
        IPC::handleMessage<Messages::NetworkProcess::DisableServiceWorkerProcessTerminationDelay>(decoder, this, &NetworkProcess::disableServiceWorkerProcessTerminationDelay);
        return;
    }
#endif
#if ENABLE(INDEXED_DATABASE)
    if (decoder.messageName() == Messages::NetworkProcess::SetIDBPerOriginQuota::name()) {
        IPC::handleMessage<Messages::NetworkProcess::SetIDBPerOriginQuota>(decoder, this, &NetworkProcess::setIDBPerOriginQuota);
        return;
    }
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

void NetworkProcess::didReceiveSyncNetworkProcessMessage(IPC::Connection& connection, IPC::Decoder& decoder, std::unique_ptr<IPC::Encoder>& replyEncoder)
{
    if (decoder.messageName() == Messages::NetworkProcess::SetAllowsAnySSLCertificateForWebSocket::name()) {
        IPC::handleMessageDelayed<Messages::NetworkProcess::SetAllowsAnySSLCertificateForWebSocket>(connection, decoder, replyEncoder, this, &NetworkProcess::setAllowsAnySSLCertificateForWebSocket);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcess::ProcessWillSuspendImminently::name()) {
        IPC::handleMessageDelayed<Messages::NetworkProcess::ProcessWillSuspendImminently>(connection, decoder, replyEncoder, this, &NetworkProcess::processWillSuspendImminently);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

