/*
 * Copyright (C) 2010-2018 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include "WebPasteboardProxy.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#if PLATFORM(COCOA)
#include "SandboxExtension.h"
#endif
#if PLATFORM(IOS_FAMILY) || PLATFORM(COCOA)
#include "SharedMemory.h"
#endif
#include "WebCoreArgumentCoders.h"
#include "WebPasteboardProxyMessages.h"
#if PLATFORM(GTK)
#include "WebSelectionData.h"
#endif
#if PLATFORM(COCOA)
#include <WebCore/Color.h>
#endif
#include <WebCore/Pasteboard.h>
#if PLATFORM(IOS_FAMILY)
#include <WebCore/PasteboardItemInfo.h>
#endif
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace WebKit {

void WebPasteboardProxy::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteURLToPasteboard::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::WriteURLToPasteboard>(decoder, this, &WebPasteboardProxy::writeURLToPasteboard);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteWebContentToPasteboard::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::WriteWebContentToPasteboard>(decoder, this, &WebPasteboardProxy::writeWebContentToPasteboard);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteImageToPasteboard::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::WriteImageToPasteboard>(decoder, this, &WebPasteboardProxy::writeImageToPasteboard);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteStringToPasteboard::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::WriteStringToPasteboard>(decoder, this, &WebPasteboardProxy::writeStringToPasteboard);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::UpdateSupportedTypeIdentifiers::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::UpdateSupportedTypeIdentifiers>(decoder, this, &WebPasteboardProxy::updateSupportedTypeIdentifiers);
        return;
    }
#endif
#if PLATFORM(GTK)
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteToClipboard::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::WriteToClipboard>(decoder, this, &WebPasteboardProxy::writeToClipboard);
        return;
    }
#endif
#if USE(LIBWPE)
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteWebContentToPasteboard::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::WriteWebContentToPasteboard>(decoder, this, &WebPasteboardProxy::writeWebContentToPasteboard);
        return;
    }
#endif
#if USE(LIBWPE)
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteStringToPasteboard::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::WriteStringToPasteboard>(decoder, this, &WebPasteboardProxy::writeStringToPasteboard);
        return;
    }
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

void WebPasteboardProxy::didReceiveSyncMessage(IPC::Connection& connection, IPC::Decoder& decoder, std::unique_ptr<IPC::Encoder>& replyEncoder)
{
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::ReadStringFromPasteboard::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::ReadStringFromPasteboard>(decoder, *replyEncoder, this, &WebPasteboardProxy::readStringFromPasteboard);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::ReadURLFromPasteboard::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::ReadURLFromPasteboard>(decoder, *replyEncoder, this, &WebPasteboardProxy::readURLFromPasteboard);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::ReadBufferFromPasteboard::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::ReadBufferFromPasteboard>(decoder, *replyEncoder, this, &WebPasteboardProxy::readBufferFromPasteboard);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardItemsCount::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::GetPasteboardItemsCount>(decoder, *replyEncoder, this, &WebPasteboardProxy::getPasteboardItemsCount);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::AllPasteboardItemInfo::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::AllPasteboardItemInfo>(decoder, *replyEncoder, this, &WebPasteboardProxy::allPasteboardItemInfo);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::InformationForItemAtIndex::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::InformationForItemAtIndex>(decoder, *replyEncoder, this, &WebPasteboardProxy::informationForItemAtIndex);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardTypesByFidelityForItemAtIndex::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::GetPasteboardTypesByFidelityForItemAtIndex>(decoder, *replyEncoder, this, &WebPasteboardProxy::getPasteboardTypesByFidelityForItemAtIndex);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteCustomData::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::WriteCustomData>(decoder, *replyEncoder, this, &WebPasteboardProxy::writeCustomData);
        return;
    }
    if (decoder.messageName() == Messages::WebPasteboardProxy::TypesSafeForDOMToReadAndWrite::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::TypesSafeForDOMToReadAndWrite>(decoder, *replyEncoder, this, &WebPasteboardProxy::typesSafeForDOMToReadAndWrite);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetNumberOfFiles::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::GetNumberOfFiles>(decoder, *replyEncoder, this, &WebPasteboardProxy::getNumberOfFiles);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardTypes::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::GetPasteboardTypes>(decoder, *replyEncoder, this, &WebPasteboardProxy::getPasteboardTypes);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardPathnamesForType::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::GetPasteboardPathnamesForType>(connection, decoder, *replyEncoder, this, &WebPasteboardProxy::getPasteboardPathnamesForType);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardStringForType::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::GetPasteboardStringForType>(decoder, *replyEncoder, this, &WebPasteboardProxy::getPasteboardStringForType);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardStringsForType::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::GetPasteboardStringsForType>(decoder, *replyEncoder, this, &WebPasteboardProxy::getPasteboardStringsForType);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardBufferForType::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::GetPasteboardBufferForType>(decoder, *replyEncoder, this, &WebPasteboardProxy::getPasteboardBufferForType);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::PasteboardCopy::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::PasteboardCopy>(decoder, *replyEncoder, this, &WebPasteboardProxy::pasteboardCopy);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardChangeCount::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::GetPasteboardChangeCount>(decoder, *replyEncoder, this, &WebPasteboardProxy::getPasteboardChangeCount);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardUniqueName::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::GetPasteboardUniqueName>(decoder, *replyEncoder, this, &WebPasteboardProxy::getPasteboardUniqueName);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardColor::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::GetPasteboardColor>(decoder, *replyEncoder, this, &WebPasteboardProxy::getPasteboardColor);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardURL::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::GetPasteboardURL>(decoder, *replyEncoder, this, &WebPasteboardProxy::getPasteboardURL);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::AddPasteboardTypes::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::AddPasteboardTypes>(decoder, *replyEncoder, this, &WebPasteboardProxy::addPasteboardTypes);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::SetPasteboardTypes::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::SetPasteboardTypes>(decoder, *replyEncoder, this, &WebPasteboardProxy::setPasteboardTypes);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::SetPasteboardURL::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::SetPasteboardURL>(connection, decoder, *replyEncoder, this, &WebPasteboardProxy::setPasteboardURL);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::SetPasteboardColor::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::SetPasteboardColor>(decoder, *replyEncoder, this, &WebPasteboardProxy::setPasteboardColor);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::SetPasteboardStringForType::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::SetPasteboardStringForType>(decoder, *replyEncoder, this, &WebPasteboardProxy::setPasteboardStringForType);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::SetPasteboardBufferForType::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::SetPasteboardBufferForType>(decoder, *replyEncoder, this, &WebPasteboardProxy::setPasteboardBufferForType);
        return;
    }
#endif
#if PLATFORM(GTK)
    if (decoder.messageName() == Messages::WebPasteboardProxy::ReadFromClipboard::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::ReadFromClipboard>(decoder, *replyEncoder, this, &WebPasteboardProxy::readFromClipboard);
        return;
    }
#endif
#if USE(LIBWPE)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardTypes::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::GetPasteboardTypes>(decoder, *replyEncoder, this, &WebPasteboardProxy::getPasteboardTypes);
        return;
    }
#endif
#if USE(LIBWPE)
    if (decoder.messageName() == Messages::WebPasteboardProxy::ReadStringFromPasteboard::name()) {
        IPC::handleMessageLegacySync<Messages::WebPasteboardProxy::ReadStringFromPasteboard>(decoder, *replyEncoder, this, &WebPasteboardProxy::readStringFromPasteboard);
        return;
    }
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

