/*
 * Copyright (C) 2010-2018 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include "WebProcessProxy.h"

#include "ArgumentCoders.h"
#include "Attachment.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "SessionState.h"
#include "WebCoreArgumentCoders.h"
#include "WebProcessProxyMessages.h"
#include <WebCore/MessagePortIdentifier.h>
#include <WebCore/MessageWithMessagePorts.h>
#if ENABLE(NETSCAPE_PLUGIN_API)
#include <WebCore/PluginData.h>
#endif
#include <WebCore/PrewarmInformation.h>
#if ENABLE(NETSCAPE_PLUGIN_API)
#include <wtf/Optional.h>
#endif
#if ENABLE(NETSCAPE_PLUGIN_API) || PLATFORM(COCOA)
#include <wtf/Vector.h>
#endif
#include <wtf/text/WTFString.h>

namespace Messages {

namespace WebProcessProxy {

#if ENABLE(NETSCAPE_PLUGIN_API)

void GetPluginProcessConnection::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const IPC::Attachment& connectionHandle, bool supportsAsynchronousInitialization)
{
    *encoder << connectionHandle;
    *encoder << supportsAsynchronousInitialization;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

void GetNetworkProcessConnection::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const IPC::Attachment& connectionHandle)
{
    *encoder << connectionHandle;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace WebProcessProxy

} // namespace Messages

namespace WebKit {

void WebProcessProxy::didReceiveWebProcessProxyMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    if (decoder.messageName() == Messages::WebProcessProxy::UpdateBackForwardItem::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::UpdateBackForwardItem>(decoder, this, &WebProcessProxy::updateBackForwardItem);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::DidDestroyFrame::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::DidDestroyFrame>(decoder, this, &WebProcessProxy::didDestroyFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::DidDestroyUserGestureToken::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::DidDestroyUserGestureToken>(decoder, this, &WebProcessProxy::didDestroyUserGestureToken);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::EnableSuddenTermination::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::EnableSuddenTermination>(decoder, this, &WebProcessProxy::enableSuddenTermination);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::DisableSuddenTermination::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::DisableSuddenTermination>(decoder, this, &WebProcessProxy::disableSuddenTermination);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::ProcessReadyToSuspend::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::ProcessReadyToSuspend>(decoder, this, &WebProcessProxy::processReadyToSuspend);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::DidCancelProcessSuspension::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::DidCancelProcessSuspension>(decoder, this, &WebProcessProxy::didCancelProcessSuspension);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::SetIsHoldingLockedFiles::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::SetIsHoldingLockedFiles>(decoder, this, &WebProcessProxy::setIsHoldingLockedFiles);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::DidExceedActiveMemoryLimit::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::DidExceedActiveMemoryLimit>(decoder, this, &WebProcessProxy::didExceedActiveMemoryLimit);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::DidExceedInactiveMemoryLimit::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::DidExceedInactiveMemoryLimit>(decoder, this, &WebProcessProxy::didExceedInactiveMemoryLimit);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::DidExceedCPULimit::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::DidExceedCPULimit>(decoder, this, &WebProcessProxy::didExceedCPULimit);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::StopResponsivenessTimer::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::StopResponsivenessTimer>(decoder, this, &WebProcessProxy::stopResponsivenessTimer);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::DidReceiveMainThreadPing::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::DidReceiveMainThreadPing>(decoder, this, &WebProcessProxy::didReceiveMainThreadPing);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::DidReceiveBackgroundResponsivenessPing::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::DidReceiveBackgroundResponsivenessPing>(decoder, this, &WebProcessProxy::didReceiveBackgroundResponsivenessPing);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::MemoryPressureStatusChanged::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::MemoryPressureStatusChanged>(decoder, this, &WebProcessProxy::memoryPressureStatusChanged);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::DidExceedInactiveMemoryLimitWhileActive::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::DidExceedInactiveMemoryLimitWhileActive>(decoder, this, &WebProcessProxy::didExceedInactiveMemoryLimitWhileActive);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::CreateNewMessagePortChannel::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::CreateNewMessagePortChannel>(decoder, this, &WebProcessProxy::createNewMessagePortChannel);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::EntangleLocalPortInThisProcessToRemote::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::EntangleLocalPortInThisProcessToRemote>(decoder, this, &WebProcessProxy::entangleLocalPortInThisProcessToRemote);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::MessagePortDisentangled::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::MessagePortDisentangled>(decoder, this, &WebProcessProxy::messagePortDisentangled);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::MessagePortClosed::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::MessagePortClosed>(decoder, this, &WebProcessProxy::messagePortClosed);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::TakeAllMessagesForPort::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::TakeAllMessagesForPort>(decoder, this, &WebProcessProxy::takeAllMessagesForPort);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::PostMessageToRemote::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::PostMessageToRemote>(decoder, this, &WebProcessProxy::postMessageToRemote);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::CheckRemotePortForActivity::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::CheckRemotePortForActivity>(decoder, this, &WebProcessProxy::checkRemotePortForActivity);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::DidDeliverMessagePortMessages::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::DidDeliverMessagePortMessages>(decoder, this, &WebProcessProxy::didDeliverMessagePortMessages);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::DidCheckProcessLocalPortForActivity::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::DidCheckProcessLocalPortForActivity>(decoder, this, &WebProcessProxy::didCheckProcessLocalPortForActivity);
        return;
    }
    if (decoder.messageName() == Messages::WebProcessProxy::DidCollectPrewarmInformation::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::DidCollectPrewarmInformation>(decoder, this, &WebProcessProxy::didCollectPrewarmInformation);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebProcessProxy::CacheMediaMIMETypes::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::CacheMediaMIMETypes>(decoder, this, &WebProcessProxy::cacheMediaMIMETypes);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebProcessProxy::RequestHighPerformanceGPU::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::RequestHighPerformanceGPU>(decoder, this, &WebProcessProxy::requestHighPerformanceGPU);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebProcessProxy::ReleaseHighPerformanceGPU::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::ReleaseHighPerformanceGPU>(decoder, this, &WebProcessProxy::releaseHighPerformanceGPU);
        return;
    }
#endif
#if PLATFORM(MAC) && ENABLE(WEBPROCESS_WINDOWSERVER_BLOCKING)
    if (decoder.messageName() == Messages::WebProcessProxy::StartDisplayLink::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::StartDisplayLink>(decoder, this, &WebProcessProxy::startDisplayLink);
        return;
    }
#endif
#if PLATFORM(MAC) && ENABLE(WEBPROCESS_WINDOWSERVER_BLOCKING)
    if (decoder.messageName() == Messages::WebProcessProxy::StopDisplayLink::name()) {
        IPC::handleMessage<Messages::WebProcessProxy::StopDisplayLink>(decoder, this, &WebProcessProxy::stopDisplayLink);
        return;
    }
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

void WebProcessProxy::didReceiveSyncWebProcessProxyMessage(IPC::Connection& connection, IPC::Decoder& decoder, std::unique_ptr<IPC::Encoder>& replyEncoder)
{
    if (decoder.messageName() == Messages::WebProcessProxy::ShouldTerminate::name()) {
        IPC::handleMessageLegacySync<Messages::WebProcessProxy::ShouldTerminate>(decoder, *replyEncoder, this, &WebProcessProxy::shouldTerminate);
        return;
    }
#if ENABLE(NETSCAPE_PLUGIN_API)
    if (decoder.messageName() == Messages::WebProcessProxy::GetPlugins::name()) {
        IPC::handleMessageLegacySync<Messages::WebProcessProxy::GetPlugins>(decoder, *replyEncoder, this, &WebProcessProxy::getPlugins);
        return;
    }
#endif
#if ENABLE(NETSCAPE_PLUGIN_API)
    if (decoder.messageName() == Messages::WebProcessProxy::GetPluginProcessConnection::name()) {
        IPC::handleMessageDelayed<Messages::WebProcessProxy::GetPluginProcessConnection>(connection, decoder, replyEncoder, this, &WebProcessProxy::getPluginProcessConnection);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebProcessProxy::GetNetworkProcessConnection::name()) {
        IPC::handleMessageDelayed<Messages::WebProcessProxy::GetNetworkProcessConnection>(connection, decoder, replyEncoder, this, &WebProcessProxy::getNetworkProcessConnection);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

